@@ -35,6 +35,7 @@
 #include <errno.h>
 #include <termios.h>
 #include <sys/ioctl.h>
+#include "config.h"
 
 #if (ULONG_MAX == 4294967295UL)
 #define MEMTEST_32BIT
@@ -78,10 +79,8 @@ void memtest_progress_end(void) {
 void memtest_progress_step(size_t curr, size_t size, char c) {
     size_t chars = ((unsigned long long)curr*progress_full)/size, j;
 
-    for (j = 0; j < chars-progress_printed; j++) {
-        printf("%c",c);
-        progress_printed++;
-    }
+    for (j = 0; j < chars-progress_printed; j++) printf("%c",c);
+    progress_printed = chars;
     fflush(stdout);
 }
 
@@ -240,6 +239,32 @@ void memtest_test(size_t megabytes, int passes) {
     }
 }
 
+void memtest_non_destructive_invert(void *addr, size_t size) {
+    volatile unsigned long *p = addr;
+    size_t words = size / sizeof(unsigned long);
+    size_t j;
+
+    /* Invert */
+    for (j = 0; j < words; j++)
+        p[j] = ~p[j];
+}
+
+void memtest_non_destructive_swap(void *addr, size_t size) {
+    volatile unsigned long *p = addr;
+    size_t words = size / sizeof(unsigned long);
+    size_t j;
+
+    /* Swap */
+    for (j = 0; j < words; j += 2) {
+        unsigned long a, b;
+
+        a = p[j];
+        b = p[j+1];
+        p[j] = b;
+        p[j+1] = a;
+    }
+}
+
 void memtest(size_t megabytes, int passes) {
     if (ioctl(1, TIOCGWINSZ, &ws) == -1) {
         ws.ws_col = 80;
