@@ -1,15 +1,27 @@
+start_server {tags {"auth"}} {
+    test {AUTH fails if there is no password configured server side} {
+        catch {r auth foo} err
+        set _ $err
+    } {ERR*no password*}
+}
+
 start_server {tags {"auth"} overrides {requirepass foobar}} {
     test {AUTH fails when a wrong password is given} {
         catch {r auth wrong!} err
-        format $err
+        set _ $err
     } {ERR*invalid password}
     
     test {Arbitrary command gives an error when AUTH is required} {
         catch {r set foo bar} err
-        format $err
-    } {ERR*operation not permitted}
+        set _ $err
+    } {NOAUTH*}
 
     test {AUTH succeeds when the right password is given} {
         r auth foobar
     } {OK}
+
+    test {Once AUTH succeeded we can actually send commands to the server} {
+        r set foo 100
+        r incr foo
+    } {101}
 }
